"""Backend for DDSIM."""

from __future__ import annotations

import time
import uuid
from typing import TYPE_CHECKING, Any, Union, cast

from qiskit import QuantumCircuit
from qiskit.providers import BackendV2, Options
from qiskit.result import Result
from qiskit.result.models import ExperimentResult, ExperimentResultData
from qiskit.transpiler import Target

from . import __version__
from .header import DDSIMHeader
from .job import DDSIMJob
from .pyddsim import CircuitSimulator
from .target import DDSIMTargetBuilder

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from qiskit.circuit import Parameter
    from qiskit.circuit.parameterexpression import ParameterValueType

    Parameters = Union[Mapping[Parameter, ParameterValueType], Sequence[ParameterValueType]]


class QasmSimulatorBackend(BackendV2):  # type: ignore[misc]
    """Python interface to MQT DDSIM."""

    _SHOW_STATE_VECTOR = False
    _TARGET = Target(description="MQT DDSIM Simulator Target", num_qubits=128)

    @staticmethod
    def _add_operations_to_target(target: Target) -> None:
        DDSIMTargetBuilder.add_0q_gates(target)
        DDSIMTargetBuilder.add_1q_gates(target)
        DDSIMTargetBuilder.add_2q_gates(target)
        DDSIMTargetBuilder.add_3q_gates(target)
        DDSIMTargetBuilder.add_multi_qubit_gates(target)
        DDSIMTargetBuilder.add_non_unitary_operations(target)
        DDSIMTargetBuilder.add_barrier(target)

    def _initialize_target(self) -> None:
        if len(self.target.operations) > 0:
            return
        self._add_operations_to_target(self.target)

    def __init__(
        self,
        name: str = "qasm_simulator",
        description: str = "MQT DDSIM QASM Simulator",
    ) -> None:
        """Constructor for the DDSIM QASM simulator backend."""
        super().__init__(name=name, description=description, backend_version=__version__)
        self._initialize_target()

    @classmethod
    def _default_options(cls) -> Options:
        return Options(
            shots=1024,
            memory=False,
            parameter_binds=None,
            seed_simulator=None,
            approximation_step_fidelity=1.0,
            approximation_steps=1,
            approximation_strategy="fidelity",
        )

    @property
    def target(self) -> Target:
        """Return the target of the backend."""
        return self._TARGET

    @property
    def max_circuits(self) -> int | None:
        """Return the maximum number of circuits that can be run in a single job."""
        return None

    @staticmethod
    def assign_parameters(
        quantum_circuits: Sequence[QuantumCircuit],
        parameter_values: Sequence[Parameters] | None,
    ) -> list[QuantumCircuit]:
        """Assign parameter values to the circuits.

        Args:
            quantum_circuits: The quantum circuits to assign parameters to.
            parameter_values: The parameter values to bind to the circuits.

        Returns:
            The bound circuits.

        Raises:
            ValueError: If the number of circuits does not match the number of provided parameter sets.
        """
        if not any(qc.parameters for qc in quantum_circuits) and not parameter_values:
            return list(quantum_circuits)

        if parameter_values is None:
            msg = "No parameter values provided although at least one parameterized circuit was supplied."
            raise ValueError(msg)

        if len(quantum_circuits) != len(parameter_values):
            msg = f"The number of circuits ({len(quantum_circuits)}) does not match the number of provided parameter sets ({len(parameter_values)})."
            raise ValueError(msg)

        bound_circuits = [
            qc.assign_parameters(parameters=values) for qc, values in zip(quantum_circuits, parameter_values)
        ]

        # fix the circuit names
        for qcb, qc in zip(bound_circuits, quantum_circuits):
            qcb.name = qc.name

        return bound_circuits

    def run(
        self,
        quantum_circuits: QuantumCircuit | Sequence[QuantumCircuit],
        parameter_values: Sequence[Parameters] | None = None,
        **options: Any,
    ) -> DDSIMJob:
        """Run a quantum circuit or list of quantum circuits on the DDSIM backend.

        Args:
            quantum_circuits: The quantum circuit(s) to run.
            parameter_values: The parameter values to bind to the circuits.
            options: Additional run options.

        Returns:
            The DDSIM job
        """
        if isinstance(quantum_circuits, QuantumCircuit):
            quantum_circuits = [quantum_circuits]

        job_id = str(uuid.uuid4())
        local_job = DDSIMJob(self, job_id, self._run_job, quantum_circuits, parameter_values, **options)
        local_job.submit()
        return local_job

    def _validate(self, quantum_circuits: Sequence[QuantumCircuit]) -> None:
        pass

    def _run_job(
        self,
        job_id: int,
        quantum_circuits: Sequence[QuantumCircuit],
        parameter_values: Sequence[Parameters] | None,
        **options: Any,
    ) -> Result:
        self._validate(quantum_circuits)
        start = time.time()

        bound_circuits = self.assign_parameters(quantum_circuits, parameter_values)
        result_list = [self._run_experiment(q_circ, **options) for q_circ in bound_circuits]

        end = time.time()

        return Result(
            backend_name=self.name,
            backend_version=self.backend_version,
            qobj_id=None,
            job_id=job_id,
            success=all(res.success for res in result_list),
            results=result_list,
            status="COMPLETED",
            time_taken=end - start,
        )

    def _run_experiment(self, qc: QuantumCircuit, **options: dict[str, Any]) -> ExperimentResult:
        start_time = time.time()
        approximation_step_fidelity = cast(float, options.get("approximation_step_fidelity", 1.0))
        approximation_steps = cast(int, options.get("approximation_steps", 1))
        approximation_strategy = str(options.get("approximation_strategy", "fidelity"))
        seed = cast(int, options.get("seed_simulator", -1))
        shots = cast(int, options.get("shots", 1024))

        sim = CircuitSimulator(
            qc,
            approximation_step_fidelity=approximation_step_fidelity,
            approximation_steps=approximation_steps,
            approximation_strategy=approximation_strategy,
            seed=seed,
        )
        counts = sim.simulate(shots=shots)
        end_time = time.time()

        data = ExperimentResultData(
            counts={hex(int(result, 2)): count for result, count in counts.items()},
            statevector=None if not self._SHOW_STATE_VECTOR else sim.get_vector(),
            time_taken=end_time - start_time,
        )

        return ExperimentResult(
            shots=shots,
            success=True,
            status="DONE",
            seed=seed,
            data=data,
            metadata=qc.metadata,
            header=DDSIMHeader(qc),
        )



"""Backend for DDSIM."""

from __future__ import annotations

from qiskit.transpiler import Target

from .qasmsimulator import QasmSimulatorBackend


class StatevectorSimulatorBackend(QasmSimulatorBackend):
    """Python interface to MQT DDSIM."""

    _SHOW_STATE_VECTOR = True
    _SV_TARGET = Target(
        description="MQT DDSIM Statevector Simulator Target",
        num_qubits=30,  # corresponds to 16GiB memory for storing the full statevector
    )

    def __init__(self) -> None:
        """Constructor for the DDSIM Statevector simulator backend."""
        super().__init__(name="statevector_simulator", description="MQT DDSIM Statevector Simulator")

    @property
    def target(self) -> Target:
        """Return the target of the backend."""
        return self._SV_TARGET



"""MQT DDSim Python Package."""

from __future__ import annotations

from ._version import version as __version__
from .provider import DDSIMProvider
from .pyddsim import (
    CircuitSimulator,
    ConstructionMode,
    DeterministicNoiseSimulator,
    HybridCircuitSimulator,
    HybridMode,
    PathCircuitSimulator,
    PathSimulatorConfiguration,
    PathSimulatorMode,
    StochasticNoiseSimulator,
    UnitarySimulator,
    dump_tensor_network,
    get_matrix,
)

__all__ = [
    "CircuitSimulator",
    "ConstructionMode",
    "DDSIMProvider",
    "DeterministicNoiseSimulator",
    "HybridCircuitSimulator",
    "HybridMode",
    "PathCircuitSimulator",
    "PathSimulatorConfiguration",
    "PathSimulatorMode",
    "StochasticNoiseSimulator",
    "UnitarySimulator",
    "__version__",
    "dump_tensor_network",
    "get_matrix",
]



"""Backend for DDSIM Hybrid Schrodinger-Feynman Simulator."""

from __future__ import annotations

from qiskit.transpiler import Target

from .hybridqasmsimulator import HybridQasmSimulatorBackend


class HybridStatevectorSimulatorBackend(HybridQasmSimulatorBackend):
    """Python interface to MQT DDSIM Hybrid Schrodinger-Feynman Simulator."""

    _SHOW_STATE_VECTOR = True
    _HSF_SV_TARGET = Target(
        description="MQT DDSIM HSF Statevector Simulator Target",
        num_qubits=30,  # corresponds to 16GiB memory for storing the full statevector
    )

    def __init__(self) -> None:
        """Constructor for the DDSIM Hybrid Schrodinger-Feynman Statevector simulator backend."""
        super().__init__(
            name="hybrid_statevector_simulator",
            description="MQT DDSIM Hybrid Schrodinger-Feynman Statevector simulator",
        )

    @property
    def target(self) -> Target:
        """Return the target of the backend."""
        return self._HSF_SV_TARGET



"""Backend for DDSIM Unitary Simulator."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any

import numpy as np
import numpy.typing as npt
from qiskit import QiskitError
from qiskit.providers import Options
from qiskit.result.models import ExperimentResult, ExperimentResultData
from qiskit.transpiler import Target

from .header import DDSIMHeader
from .pyddsim import ConstructionMode, UnitarySimulator, get_matrix
from .qasmsimulator import QasmSimulatorBackend
from .target import DDSIMTargetBuilder

if TYPE_CHECKING:
    from collections.abc import Sequence

    from qiskit import QuantumCircuit


class UnitarySimulatorBackend(QasmSimulatorBackend):
    """Decision diagram-based unitary simulator."""

    _US_TARGET = Target(
        description="MQT DDSIM Unitary Simulator Target",
        num_qubits=15,  # corresponds to 16GiB memory for storing the full matrix
    )

    @staticmethod
    def _add_operations_to_target(target: Target) -> None:
        DDSIMTargetBuilder.add_0q_gates(target)
        DDSIMTargetBuilder.add_1q_gates(target)
        DDSIMTargetBuilder.add_2q_gates(target)
        DDSIMTargetBuilder.add_3q_gates(target)
        DDSIMTargetBuilder.add_multi_qubit_gates(target)
        DDSIMTargetBuilder.add_barrier(target)

    def __init__(self) -> None:
        """Constructor for the DDSIM unitary simulator backend."""
        super().__init__(name="unitary_simulator", description="MQT DDSIM Unitary Simulator")

    @classmethod
    def _default_options(cls) -> Options:
        return Options(shots=1, mode="recursive", parameter_binds=None)

    @property
    def target(self) -> Target:
        """Return the target of the backend."""
        return self._US_TARGET

    @classmethod
    def _run_experiment(cls, qc: QuantumCircuit, **options: Any) -> ExperimentResult:
        start_time = time.time()
        seed = options.get("seed", -1)
        mode = options.get("mode", "recursive")

        if mode == "sequential":
            construction_mode = ConstructionMode.sequential
        elif mode == "recursive":
            construction_mode = ConstructionMode.recursive
        else:
            msg = (
                f"Construction mode {mode} not supported by DDSIM unitary simulator. Available modes are "
                "'recursive' and 'sequential'"
            )
            raise QiskitError(msg)

        sim = UnitarySimulator(qc, seed=seed, mode=construction_mode)
        sim.construct()
        # Extract resulting matrix from final DD and write data
        unitary: npt.NDArray[np.complex128] = np.zeros((2**qc.num_qubits, 2**qc.num_qubits), dtype=np.complex128)
        get_matrix(sim, unitary)
        end_time = time.time()

        data = ExperimentResultData(
            unitary=unitary,
            construction_time=sim.get_construction_time(),
            max_dd_nodes=sim.get_max_node_count(),
            dd_nodes=sim.get_final_node_count(),
            time_taken=end_time - start_time,
        )

        return ExperimentResult(
            shots=1,
            success=True,
            status="DONE",
            seed=seed,
            data=data,
            metadata=qc.metadata,
            header=DDSIMHeader(qc),
        )

    def _validate(self, quantum_circuits: Sequence[QuantumCircuit]) -> None:
        """Semantic validations of the quantum circuits which cannot be done via schemas.

        1. No shots
        2. No measurements in the middle.
        """
        for qc in quantum_circuits:
            name = qc.name
            n_qubits = qc.num_qubits
            max_qubits = self.target.num_qubits

            if n_qubits > max_qubits:
                msg = f"Number of qubits {n_qubits} is greater than maximum ({max_qubits}) for '{self.name}'."
                raise QiskitError(msg)

            if qc.metadata is not None and "shots" in qc.metadata and qc.metadata["shots"] != 1:
                qc.metadata["shots"] = 1

            for instruction in qc.data:
                operation = instruction.operation
                if operation.name in {"measure", "reset"}:
                    msg = f"Unsupported '{self.name}' instruction '{operation.name}' in circuit '{name}'."
                    raise QiskitError(msg)



from typing import Any, ClassVar, overload

import numpy as np
from numpy.typing import NDArray
from qiskit import QuantumCircuit

__all__ = [
    "CircuitSimulator",
    "ConstructionMode",
    "DeterministicNoiseSimulator",
    "HybridCircuitSimulator",
    "HybridMode",
    "PathCircuitSimulator",
    "PathSimulatorConfiguration",
    "PathSimulatorMode",
    "StochasticNoiseSimulator",
    "UnitarySimulator",
    "dump_tensor_network",
    "get_matrix",
]

class CircuitSimulator:
    def __init__(
        self,
        circ: QuantumCircuit | str,
        approximation_step_fidelity: float = 1.0,
        approximation_steps: int = 1,
        approximation_strategy: str = "fidelity",
        seed: int = -1,
    ) -> None: ...
    def expectation_value(self, observable: QuantumCircuit | str) -> float: ...
    def export_dd_to_graphviz_file(
        self,
        filename: str,
        colored: bool = True,
        edge_labels: bool = False,
        classic: bool = False,
        memory: bool = False,
        format_as_polar: bool = True,
    ) -> None: ...
    def export_dd_to_graphviz_str(
        self,
        colored: bool = True,
        edge_labels: bool = False,
        classic: bool = False,
        memory: bool = False,
        format_as_polar: bool = True,
    ) -> str: ...
    def get_active_matrix_node_count(self) -> int: ...
    def get_active_vector_node_count(self) -> int: ...
    def get_max_matrix_node_count(self) -> int: ...
    def get_max_vector_node_count(self) -> int: ...
    def get_name(self) -> str: ...
    def get_number_of_qubits(self) -> int: ...
    def get_tolerance(self) -> float: ...
    def get_vector(self) -> list[complex]: ...
    def set_tolerance(self, tol: float) -> None: ...
    def simulate(self, shots: int) -> dict[str, int]: ...
    def statistics(self) -> dict[str, str]: ...

class DeterministicNoiseSimulator:
    def __init__(
        self,
        circ: QuantumCircuit | str,
        approximation_step_fidelity: float = 1.0,
        approximation_steps: int = 1,
        approximation_strategy: str = "fidelity",
        seed: int = -1,
        noise_effects: str = "APD",
        noise_probability: float = 0.01,
        amp_damping_probability: float | None = 0.02,
        multi_qubit_gate_factor: float = 2,
    ) -> None: ...
    def export_dd_to_graphviz_file(
        self,
        filename: str,
        colored: bool = True,
        edge_labels: bool = False,
        classic: bool = False,
        memory: bool = False,
        format_as_polar: bool = True,
    ) -> None: ...
    def export_dd_to_graphviz_str(
        self,
        colored: bool = True,
        edge_labels: bool = False,
        classic: bool = False,
        memory: bool = False,
        format_as_polar: bool = True,
    ) -> str: ...
    def get_active_matrix_node_count(self) -> int: ...
    def get_active_vector_node_count(self) -> int: ...
    def get_max_matrix_node_count(self) -> int: ...
    def get_max_vector_node_count(self) -> int: ...
    def get_name(self) -> str: ...
    def get_number_of_qubits(self) -> int: ...
    def get_tolerance(self) -> float: ...
    def get_vector(self) -> list[complex]: ...
    def set_tolerance(self, tol: float) -> None: ...
    def simulate(self, shots: int) -> dict[str, int]: ...
    def statistics(self) -> dict[str, str]: ...

class HybridMode:
    DD: ClassVar[HybridMode]  # value = <HybridMode.DD: 0>
    __members__: ClassVar[
        dict[str, HybridMode]
    ]  # value = {'DD': <HybridMode.DD: 0>, 'amplitude': <HybridMode.amplitude: 1>}
    amplitude: ClassVar[HybridMode]  # value = <HybridMode.amplitude: 1>

    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class HybridCircuitSimulator:
    def __init__(
        self,
        circ: QuantumCircuit | str,
        approximation_step_fidelity: float = 1.0,
        approximation_steps: int = 1,
        approximation_strategy: str = "fidelity",
        seed: int = -1,
        mode: HybridMode = ...,
        nthreads: int = 2,
    ) -> None: ...
    def export_dd_to_graphviz_file(
        self,
        filename: str,
        colored: bool = True,
        edge_labels: bool = False,
        classic: bool = False,
        memory: bool = False,
        format_as_polar: bool = True,
    ) -> None: ...
    def export_dd_to_graphviz_str(
        self,
        colored: bool = True,
        edge_labels: bool = False,
        classic: bool = False,
        memory: bool = False,
        format_as_polar: bool = True,
    ) -> str: ...
    def get_active_matrix_node_count(self) -> int: ...
    def get_active_vector_node_count(self) -> int: ...
    def get_final_amplitudes(self) -> list[complex]: ...
    def get_max_matrix_node_count(self) -> int: ...
    def get_max_vector_node_count(self) -> int: ...
    def get_mode(self) -> HybridMode: ...
    def get_name(self) -> str: ...
    def get_number_of_qubits(self) -> int: ...
    def get_tolerance(self) -> float: ...
    def get_vector(self) -> list[complex]: ...
    def set_tolerance(self, tol: float) -> None: ...
    def simulate(self, shots: int) -> dict[str, int]: ...
    def statistics(self) -> dict[str, str]: ...

class PathSimulatorMode:
    __members__: ClassVar[
        dict[str, PathSimulatorMode]
    ]  # value = {'sequential': <PathSimulatorMode.sequential: 0>, 'pairwise_recursive': <PathSimulatorMode.pairwise_recursive: 1>, 'cotengra': <PathSimulatorMode.cotengra: 4>, 'bracket': <PathSimulatorMode.bracket: 2>, 'alternating': <PathSimulatorMode.alternating: 3>, 'gate_cost': <PathSimulatorMode.gate_cost: 5>}
    alternating: ClassVar[PathSimulatorMode]  # value = <PathSimulatorMode.alternating: 3>
    bracket: ClassVar[PathSimulatorMode]  # value = <PathSimulatorMode.bracket: 2>
    cotengra: ClassVar[PathSimulatorMode]  # value = <PathSimulatorMode.cotengra: 4>
    gate_cost: ClassVar[PathSimulatorMode]  # value = <PathSimulatorMode.gate_cost: 5>
    pairwise_recursive: ClassVar[PathSimulatorMode]  # value = <PathSimulatorMode.pairwise_recursive: 1>
    sequential: ClassVar[PathSimulatorMode]  # value = <PathSimulatorMode.sequential: 0>

    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    @overload
    def __init__(self, value: int) -> None: ...
    @overload
    def __init__(self, arg0: str) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class PathSimulatorConfiguration:
    def __init__(self) -> None: ...
    def json(self) -> dict[str, Any]: ...
    @property
    def bracket_size(self) -> int: ...
    @bracket_size.setter
    def bracket_size(self, arg0: int) -> None: ...
    @property
    def gate_cost(self) -> list[int]: ...
    @gate_cost.setter
    def gate_cost(self, arg0: list[int]) -> None: ...
    @property
    def mode(self) -> PathSimulatorMode: ...
    @mode.setter
    def mode(self, arg0: PathSimulatorMode) -> None: ...
    @property
    def seed(self) -> int: ...
    @seed.setter
    def seed(self, arg0: int) -> None: ...
    @property
    def starting_point(self) -> int: ...
    @starting_point.setter
    def starting_point(self, arg0: int) -> None: ...

class PathCircuitSimulator:
    @overload
    def __init__(self, circ: QuantumCircuit | str, config: PathSimulatorConfiguration = ...) -> None: ...
    @overload
    def __init__(
        self,
        circ: QuantumCircuit | str,
        mode: PathSimulatorMode = ...,
        bracket_size: int = 2,
        starting_point: int = 0,
        gate_cost: list[int] = ...,
        seed: int = ...,
    ) -> None: ...
    def export_dd_to_graphviz_file(
        self,
        filename: str,
        colored: bool = True,
        edge_labels: bool = False,
        classic: bool = False,
        memory: bool = False,
        format_as_polar: bool = True,
    ) -> None: ...
    def export_dd_to_graphviz_str(
        self,
        colored: bool = True,
        edge_labels: bool = False,
        classic: bool = False,
        memory: bool = False,
        format_as_polar: bool = True,
    ) -> str: ...
    def get_active_matrix_node_count(self) -> int: ...
    def get_active_vector_node_count(self) -> int: ...
    def get_max_matrix_node_count(self) -> int: ...
    def get_max_vector_node_count(self) -> int: ...
    def get_name(self) -> str: ...
    def get_number_of_qubits(self) -> int: ...
    def get_tolerance(self) -> float: ...
    def get_vector(self) -> list[complex]: ...
    def set_simulation_path(self, path: list[tuple[int, int]], assume_correct_order: bool = False) -> None: ...
    def set_tolerance(self, tol: float) -> None: ...
    def simulate(self, shots: int) -> dict[str, int]: ...
    def statistics(self) -> dict[str, str]: ...

class StochasticNoiseSimulator:
    def __init__(
        self,
        circ: QuantumCircuit | str,
        approximation_step_fidelity: float = 1.0,
        approximation_steps: int = 1,
        approximation_strategy: str = "fidelity",
        seed: int = -1,
        noise_effects: str = "APD",
        noise_probability: float = 0.01,
        amp_damping_probability: float | None = 0.02,
        multi_qubit_gate_factor: float = 2,
    ) -> None: ...
    def export_dd_to_graphviz_file(
        self,
        filename: str,
        colored: bool = True,
        edge_labels: bool = False,
        classic: bool = False,
        memory: bool = False,
        format_as_polar: bool = True,
    ) -> None: ...
    def export_dd_to_graphviz_str(
        self,
        colored: bool = True,
        edge_labels: bool = False,
        classic: bool = False,
        memory: bool = False,
        format_as_polar: bool = True,
    ) -> str: ...
    def get_active_matrix_node_count(self) -> int: ...
    def get_active_vector_node_count(self) -> int: ...
    def get_max_matrix_node_count(self) -> int: ...
    def get_max_vector_node_count(self) -> int: ...
    def get_name(self) -> str: ...
    def get_number_of_qubits(self) -> int: ...
    def get_tolerance(self) -> float: ...
    def get_vector(self) -> list[complex]: ...
    def set_tolerance(self, tol: float) -> None: ...
    def simulate(self, shots: int) -> dict[str, int]: ...
    def statistics(self) -> dict[str, str]: ...

class ConstructionMode:
    __members__: ClassVar[
        dict[str, ConstructionMode]
    ]  # value = {'recursive': <ConstructionMode.recursive: 1>, 'sequential': <ConstructionMode.sequential: 0>}
    recursive: ClassVar[ConstructionMode]  # value = <ConstructionMode.recursive: 1>
    sequential: ClassVar[ConstructionMode]  # value = <ConstructionMode.sequential: 0>

    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class UnitarySimulator:
    def __init__(
        self,
        circ: QuantumCircuit | str,
        approximation_step_fidelity: float = 1.0,
        approximation_steps: int = 1,
        approximation_strategy: str = "fidelity",
        seed: int = -1,
        mode: ConstructionMode = ...,
    ) -> None: ...
    def construct(self) -> None: ...
    def export_dd_to_graphviz_file(
        self,
        filename: str,
        colored: bool = True,
        edge_labels: bool = False,
        classic: bool = False,
        memory: bool = False,
        format_as_polar: bool = True,
    ) -> None: ...
    def export_dd_to_graphviz_str(
        self,
        colored: bool = True,
        edge_labels: bool = False,
        classic: bool = False,
        memory: bool = False,
        format_as_polar: bool = True,
    ) -> str: ...
    def get_active_matrix_node_count(self) -> int: ...
    def get_active_vector_node_count(self) -> int: ...
    def get_construction_time(self) -> float: ...
    def get_final_node_count(self) -> int: ...
    def get_max_matrix_node_count(self) -> int: ...
    def get_max_node_count(self) -> int: ...
    def get_max_vector_node_count(self) -> int: ...
    def get_mode(self) -> ConstructionMode: ...
    def get_name(self) -> str: ...
    def get_number_of_qubits(self) -> int: ...
    def get_tolerance(self) -> float: ...
    def set_tolerance(self, tol: float) -> None: ...
    def statistics(self) -> dict[str, str]: ...

def dump_tensor_network(circ: QuantumCircuit | str, filename: str) -> None: ...
def get_matrix(sim: UnitarySimulator, mat: NDArray[np.complex128]) -> None: ...



"""Utilities for constructing a Qiskit experiment header for DDSIM backends."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from qiskit.result.models import QobjExperimentHeader

if TYPE_CHECKING:
    from qiskit import QuantumCircuit


@dataclass
class DDSIMHeader(QobjExperimentHeader):  # type: ignore[misc]
    """Qiskit experiment header for DDSIM backends."""

    name: str
    n_qubits: int
    memory_slots: int
    global_phase: float
    creg_sizes: list[tuple[str, int]]
    clbit_labels: list[tuple[str, int]]
    qreg_sizes: list[tuple[str, int]]
    qubit_labels: list[tuple[str, int]]

    def __init__(self, qc: QuantumCircuit) -> None:
        """Initialize a new DDSIM experiment header."""
        # no call to super class constructor because this would trigger deprecation warnings
        self.name = qc.name
        self.n_qubits = qc.num_qubits
        self.memory_slots = qc.num_clbits
        self.global_phase = qc.global_phase
        self.creg_sizes = [(creg.name, creg.size) for creg in qc.cregs]
        self.clbit_labels = [(creg.name, j) for creg in qc.cregs for j in range(creg.size)]
        self.qreg_sizes = [(qreg.name, qreg.size) for qreg in qc.qregs]
        self.qubit_labels = [(qreg.name, j) for qreg in qc.qregs for j in range(qreg.size)]



"""Backend for DDSIM."""

from __future__ import annotations

from qiskit.transpiler import Target

from .pathqasmsimulator import PathQasmSimulatorBackend


class PathStatevectorSimulatorBackend(PathQasmSimulatorBackend):
    """Python interface to MQT DDSIM Simulation Path Framework."""

    _SHOW_STATE_VECTOR = True
    _Path_SV_TARGET = Target(
        description="MQT DDSIM Simulation Path Framework Statevector Target",
        num_qubits=30,  # corresponds to 16GiB memory for storing the full statevector
    )

    def __init__(self) -> None:
        """Constructor for the DDSIM Simulation Path Framework Statevector Simulator backend."""
        super().__init__(
            name="path_sim_statevector_simulator",
            description="MQT DDSIM Simulation Path Framework Statevector Simulator",
        )

    @property
    def target(self) -> Target:
        """Return the target of the backend."""
        return self._Path_SV_TARGET



"""Provider for DDSIM backends."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

from qiskit.providers import BackendV2
from qiskit.providers.exceptions import QiskitBackendNotFoundError
from qiskit.providers.providerutils import filter_backends

from .deterministicnoisesimulator import DeterministicNoiseSimulatorBackend
from .hybridqasmsimulator import HybridQasmSimulatorBackend
from .hybridstatevectorsimulator import HybridStatevectorSimulatorBackend
from .pathqasmsimulator import PathQasmSimulatorBackend
from .pathstatevectorsimulator import PathStatevectorSimulatorBackend
from .qasmsimulator import QasmSimulatorBackend
from .statevectorsimulator import StatevectorSimulatorBackend
from .stochasticnoisesimulator import StochasticNoiseSimulatorBackend
from .unitarysimulator import UnitarySimulatorBackend

if TYPE_CHECKING:
    from collections.abc import Callable


class DDSIMProvider:
    """Provider for DDSIM backends."""

    _BACKENDS = (
        ("qasm_simulator", QasmSimulatorBackend),
        ("statevector_simulator", StatevectorSimulatorBackend),
        ("hybrid_qasm_simulator", HybridQasmSimulatorBackend),
        ("hybrid_statevector_simulator", HybridStatevectorSimulatorBackend),
        ("path_sim_qasm_simulator", PathQasmSimulatorBackend),
        ("path_sim_statevector_simulator", PathStatevectorSimulatorBackend),
        ("unitary_simulator", UnitarySimulatorBackend),
        ("stochastic_dd_simulator", StochasticNoiseSimulatorBackend),
        ("density_matrix_dd_simulator", DeterministicNoiseSimulatorBackend),
    )

    def get_backend(self, name: str | None = None, **kwargs: Any) -> BackendV2:
        """Return a backend matching the specified criteria.

        Args:
            name: Name of the backend.
            kwargs: Additional filtering criteria.
        """
        backends = self.backends(name, **kwargs)
        if len(backends) > 1:
            msg = "More than one backend matches the criteria"
            raise QiskitBackendNotFoundError(msg)
        if not backends:
            msg = "No backend matches the criteria"
            raise QiskitBackendNotFoundError(msg)

        return backends[0]

    def backends(
        self,
        name: str | None = None,
        filters: Callable[[list[BackendV2]], list[BackendV2]] | None = None,
        **kwargs: dict[str, Any],
    ) -> list[BackendV2]:
        """Return a list of backends matching the specified criteria.

        Args:
            name: Name of the backend.
            filters: Additional filtering criteria.
            kwargs: Additional filtering criteria.
        """
        backends = [
            backend_cls() for backend_name, backend_cls in self._BACKENDS if name is None or backend_name == name
        ]
        return cast(list[BackendV2], filter_backends(backends, filters=filters, **kwargs))

    def __str__(self) -> str:
        """Return the provider name."""
        return "DDSIMProvider"



"""Estimator implementation using DDSIM CircuitSimulator."""

from __future__ import annotations

from itertools import accumulate
from typing import TYPE_CHECKING, Any, Union, cast

import numpy as np
from qiskit.circuit import QuantumCircuit
from qiskit.primitives import Estimator as QiskitEstimator
from qiskit.primitives import EstimatorResult
from qiskit.quantum_info import Pauli, PauliList

from mqt.ddsim.pyddsim import CircuitSimulator
from mqt.ddsim.qasmsimulator import QasmSimulatorBackend

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from qiskit.circuit import Parameter
    from qiskit.circuit.parameterexpression import ParameterValueType

    Parameters = Union[Mapping[Parameter, ParameterValueType], Sequence[ParameterValueType]]


class Estimator(QiskitEstimator):  # type: ignore[misc]
    """DDSIM implementation of qiskit's sampler.

    Code adapted from Qiskit's BackendEstimator class.
    """

    def __init__(
        self,
        options: dict[str, Any] | None = None,
        abelian_grouping: bool = False,
    ) -> None:
        """Initialize a new Estimator instance.

        Args:
            options: Default options.
            abelian_grouping: Whether the observable should be grouped into
                commuting
        """
        super().__init__(options=options)

        self._abelian_grouping = abelian_grouping

        self._preprocessed_circuits: tuple[list[QuantumCircuit], list[list[QuantumCircuit]]] | None = None

        self._grouping = list(zip(range(len(self._circuits)), range(len(self._observables))))

    @property
    def preprocessed_circuits(
        self,
    ) -> tuple[list[QuantumCircuit], list[list[QuantumCircuit]]]:
        """Generate quantum circuits for states and observables produced by preprocessing.

        Returns:
        Tuple: A tuple containing two entries:
            - List: Quantum circuits list entered in run() method.
            - List: Quantum circuit representations of the observables.
        """
        self._preprocessed_circuits = self._preprocessing()
        return self._preprocessed_circuits

    def _preprocessing(self) -> tuple[list[QuantumCircuit], list[list[QuantumCircuit]]]:
        """Perform preprocessing for circuit arranging and packaging.

        Returns:
        Tuple:
            - List: Quantum circuits list entered in run() method.
            - List: Quantum circuit representations of the observables.

        This method performs preprocessing for circuit arranging and packaging. It processes quantum circuits and observables
        based on the specified grouping and abelian grouping conditions.
        """
        state_circuits = []
        observable_circuits = []
        for group in self._grouping:
            circuit = self._circuits[group[0]]
            observable = self._observables[group[1]]
            diff_circuits: list[QuantumCircuit] = []
            if self._abelian_grouping:
                for obs in observable.group_commuting(qubit_wise=True):
                    basis = Pauli((
                        np.logical_or.reduce(obs.paulis.z),
                        np.logical_or.reduce(obs.paulis.x),
                    ))
                    obs_circuit, indices = self._observable_circuit(circuit.num_qubits, basis)
                    paulis = PauliList.from_symplectic(
                        obs.paulis.z[:, indices],
                        obs.paulis.x[:, indices],
                        obs.paulis.phase,
                    )
                    obs_circuit.metadata = {
                        "paulis": paulis,
                        "coeffs": np.real_if_close(obs.coeffs),
                    }
                    diff_circuits.append(obs_circuit)
            else:
                for basis, obs in zip(observable.paulis, observable):
                    obs_circuit, indices = self._observable_circuit(circuit.num_qubits, basis)
                    paulis = PauliList.from_symplectic(
                        obs.paulis.z[:, indices],
                        obs.paulis.x[:, indices],
                        obs.paulis.phase,
                    )
                    obs_circuit.metadata = {
                        "paulis": paulis,
                        "coeffs": np.real_if_close(obs.coeffs),
                    }
                    diff_circuits.append(obs_circuit)

            state_circuits.append(circuit.copy())
            observable_circuits.append(diff_circuits)
        return state_circuits, observable_circuits

    @staticmethod
    def _observable_circuit(num_qubits: int, pauli: Pauli) -> tuple[QuantumCircuit, list[int]]:
        """Creates the quantum circuit representation of an observable given as a Pauli string.

        Parameters:
            - num_qubits: Number of qubits of the observable.
            - pauli_string (str): The Pauli string representing the observable.

        Returns:
        Tuple: A tuple containing two entries:
            - QuantumCircuit: The quantum circuit representation of the observable.
            - List: A list of the qubits involved in the observables operators
        """
        qubit_indices = np.arange(pauli.num_qubits)[pauli.z | pauli.x]
        if not np.any(qubit_indices):
            qubit_indices = [0]
        obs_circuit = QuantumCircuit(num_qubits, len(qubit_indices))
        for i in qubit_indices:
            if pauli.x[i]:
                if pauli.z[i]:
                    obs_circuit.y(i)
                else:
                    obs_circuit.x(i)
            elif pauli.z[i]:
                obs_circuit.z(i)

        return obs_circuit, qubit_indices

    def _call(
        self,
        circuits: Sequence[int],
        observables: Sequence[int],
        parameter_values: Sequence[Parameters],
        **run_options: dict[str, Any],
    ) -> EstimatorResult:
        # Organize circuits
        self._grouping = list(zip(circuits, observables))
        state_circuits, observable_circuits = self.preprocessed_circuits
        num_observables = [len(obs_circ_list) for obs_circ_list in observable_circuits]
        accum = [0, *list(accumulate(num_observables))]

        # Extract metadata from circuits
        metadata = [circ.metadata for obs_circ_list in observable_circuits for circ in obs_circ_list]
        for obs_circ_list in observable_circuits:
            for circ in obs_circ_list:
                circ.metadata = {}

        # Bind parameters
        bound_circuits = QasmSimulatorBackend.assign_parameters(state_circuits, parameter_values)

        # Run and bind parameters
        result_list = [
            exp
            for circ, obs_circ_list in zip(bound_circuits, observable_circuits)
            for exp in self._run_experiment(circ, obs_circ_list, **run_options)
        ]

        return self._postprocessing(result_list, accum, metadata)

    @staticmethod
    def _run_experiment(
        circ: QuantumCircuit,
        obs_circ_list: list[QuantumCircuit],
        **options: dict[str, Any],
    ) -> list[float]:
        approximation_step_fidelity = cast(float, options.get("approximation_step_fidelity", 1.0))
        approximation_steps = cast(int, options.get("approximation_steps", 1))
        approximation_strategy = str(options.get("approximation_strategy", "fidelity"))
        seed = cast(int, options.get("seed_simulator", -1))

        sim = CircuitSimulator(
            circ,
            approximation_step_fidelity=approximation_step_fidelity,
            approximation_steps=approximation_steps,
            approximation_strategy=approximation_strategy,
            seed=seed,
        )

        return [sim.expectation_value(observable=obs) for obs in obs_circ_list]

    @staticmethod
    def _postprocessing(result_list: list[float], accum: list[int], metadata: list[dict[str, Any]]) -> EstimatorResult:
        """Perform postprocessing for the evaluation of expectation values.

        Parameters:
        - result_list (list[float]): A list of measurement results.
        - accum (list[int]): A list representing accumulated indices for grouping measurement results.
        - metadata (list[dict]): A list of dictionaries containing metadata associated with each measurement.

        Returns:
        EstimatorResult: An instance of the EstimatorResult class containing the processed expectation values and metadata.

        This method calculates the expectation values by combining measurement results according to the provided accumulation indices.
        The resulting expectation values are then packaged into an EstimatorResult object along with metadata.

        The input metadata is overwritten, setting "variance" and "shots" to 0 for each expectation value.
        """
        expval_list = []

        for i, j in zip(accum, accum[1:]):
            combined_expval = 0.0

            for k in range(i, j):
                expval = [result_list[k]]
                meta = metadata[k]
                coeffs = meta["coeffs"]

                # Accumulate
                combined_expval += np.dot(expval, coeffs)

            expval_list.append(combined_expval)

        metadata = [{"variance": 0, "shots": 0} for _ in expval_list]

        return EstimatorResult(np.real_if_close(expval_list), metadata)



"""Sampler implementation using QasmSimulatorBackend."""

from __future__ import annotations

import math
from typing import TYPE_CHECKING, Any, Union

from qiskit.primitives import SamplerResult
from qiskit.primitives.sampler import Sampler as QiskitSampler
from qiskit.result import QuasiDistribution, Result

from mqt.ddsim.qasmsimulator import QasmSimulatorBackend

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from qiskit.circuit import Parameter
    from qiskit.circuit.parameterexpression import ParameterValueType

    Parameters = Union[Mapping[Parameter, ParameterValueType], Sequence[ParameterValueType]]


class Sampler(QiskitSampler):  # type: ignore[misc]
    """Sampler implementation using QasmSimulatorBackend."""

    _BACKEND = QasmSimulatorBackend()

    def __init__(
        self,
        *,
        options: dict[str, Any] | None = None,
    ) -> None:
        """Initialize a new DDSIM Sampler.

        Args:
            options: Default options.
        """
        super().__init__(options=options)

    @property
    def backend(self) -> QasmSimulatorBackend:
        """The backend used by the sampler."""
        return self._BACKEND

    @property
    def num_circuits(self) -> int:
        """The number of circuits stored in the sampler."""
        return len(self._circuits)

    def _call(
        self,
        circuits: Sequence[int],
        parameter_values: Sequence[Parameters],
        **run_options: Any,
    ) -> SamplerResult:
        """Runs DDSIM backend.

        Args:
            circuits: List of circuit indices to simulate
            parameter_values: List of parameters associated with those circuits
            run_options: Additional run options.

        Returns:
            The result of the sampling process.
        """
        result = self.backend.run([self._circuits[i] for i in circuits], parameter_values, **run_options).result()

        return self._postprocessing(result, circuits)

    @staticmethod
    def _postprocessing(result: Result, circuits: Sequence[int]) -> SamplerResult:
        """Converts counts into quasi-probability distributions.

        Args:
            result: Result from DDSIM backend
            circuits: List of circuit indices

        Returns:
            The result of the sampling process.
        """
        counts = result.get_counts()
        if not isinstance(counts, list):
            counts = [counts]

        shots = sum(counts[0].values())
        metadata: list[dict[str, Any]] = [{"shots": shots} for _ in range(len(circuits))]
        probabilities = [
            QuasiDistribution(
                {k: v / shots for k, v in count.items()},
                shots=shots,
                stddev_upper_bound=1 / math.sqrt(shots),
            )
            for count in counts
        ]

        return SamplerResult(probabilities, metadata)



"""Backend for DDSIM Stochastic Simulator."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any, cast

from qiskit.providers import Options
from qiskit.result.models import ExperimentResult, ExperimentResultData

from mqt import ddsim

from .header import DDSIMHeader
from .qasmsimulator import QasmSimulatorBackend

if TYPE_CHECKING:
    from qiskit import QuantumCircuit


class StochasticNoiseSimulatorBackend(QasmSimulatorBackend):
    """Python interface to MQT DDSIM stochastic noise-aware simulator."""

    def __init__(
        self,
        name: str = "stochastic_dd_simulator",
        description: str = "MQT DDSIM noise-aware stochastic simulator based on decision diagrams",
    ) -> None:
        """Constructor for the DDSIM stochastic simulator backend."""
        super().__init__(name=name, description=description)

    @classmethod
    def _default_options(cls) -> Options:
        return Options(
            shots=None,
            parameter_binds=None,
            simulator_seed=None,
            approximation_step_fidelity=1.0,
            approximation_steps=1,
            approximation_strategy="fidelity",
            noise_effects="APD",
            noise_probability=0.01,
            amp_damping_probability=0.02,
            multi_qubit_gate_factor=2,
        )

    @staticmethod
    def _run_experiment(qc: QuantumCircuit, **options: dict[str, Any]) -> ExperimentResult:
        start_time = time.time()
        approximation_step_fidelity = cast(float, options.get("approximation_step_fidelity", 1.0))
        approximation_steps = cast(int, options.get("approximation_steps", 1))
        approximation_strategy = cast(str, options.get("approximation_strategy", "fidelity"))
        noise_effects = cast(str, options.get("noise_effects", "APD"))
        noise_probability = cast(float, options.get("noise_probability", 0.01))
        amp_damping_probability = cast(float, options.get("amp_damping_probability", 0.02))
        multi_qubit_gate_factor = cast(float, options.get("multi_qubit_gate_factor", 2))
        seed = cast(int, options.get("seed_simulator", -1))
        shots = cast(int, options.get("shots", 1024))

        sim = ddsim.StochasticNoiseSimulator(
            circ=qc,
            approximation_step_fidelity=approximation_step_fidelity,
            approximation_steps=approximation_steps,
            approximation_strategy=approximation_strategy,
            seed=seed,
            noise_effects=noise_effects,
            noise_probability=noise_probability,
            amp_damping_probability=amp_damping_probability,
            multi_qubit_gate_factor=multi_qubit_gate_factor,
        )

        counts = sim.simulate(shots=shots)
        end_time = time.time()

        data = ExperimentResultData(
            counts={hex(int(result, 2)): count for result, count in counts.items()},
            statevector=None,
            time_taken=end_time - start_time,
        )

        return ExperimentResult(
            shots=shots,
            success=True,
            status="DONE",
            seed=seed,
            data=data,
            metadata=qc.metadata,
            header=DDSIMHeader(qc),
        )



"""Backend for DDSIM Task-Based Simulator."""

from __future__ import annotations

import pathlib
import time
from typing import TYPE_CHECKING, Any, cast

if TYPE_CHECKING:
    from quimb.tensor import Tensor, TensorNetwork

import locale

from qiskit import QuantumCircuit
from qiskit.providers import Options
from qiskit.result.models import ExperimentResult, ExperimentResultData
from qiskit.transpiler import Target

from .header import DDSIMHeader
from .pyddsim import PathCircuitSimulator, PathSimulatorConfiguration, PathSimulatorMode
from .qasmsimulator import QasmSimulatorBackend
from .target import DDSIMTargetBuilder


def read_tensor_network_file(filename: str) -> list[Tensor]:
    """Read a tensor network from a file.

    Args:
        filename: The name of the file to read the tensor network from.

    Returns:
        The tensor network read from the file.
    """
    import numpy as np
    import pandas as pd
    import quimb.tensor as qtn

    df = pd.read_json(filename)
    tensors = []

    for i in range(len(df["tensors"])):
        tens_data = [complex(real, imag) for [real, imag] in df["tensors"][i][3]]
        tens_shape = df["tensors"][i][2]
        data = np.array(tens_data).reshape(tens_shape)
        inds = df["tensors"][i][1]
        tags = df["tensors"][i][0]
        tensors.append(qtn.Tensor(data, inds, tags, left_inds=inds[: len(inds) // 2]))
    return tensors


def create_tensor_network(qc: QuantumCircuit) -> TensorNetwork:
    """Create a tensor network from a quantum circuit.

    Args:
        qc: The quantum circuit to be simulated.

    Returns:
        The tensor network representing the quantum circuit.
    """
    import quimb.tensor as qtn
    import sparse

    from mqt.ddsim import dump_tensor_network

    if isinstance(qc, QuantumCircuit):
        filename = qc.name + "_" + str(qc.num_qubits) + ".tensor"
        nqubits = qc.num_qubits
    else:
        filename = "tensor.tensor"
        nqubits = qc.header.n_qubits

    dump_tensor_network(qc, filename)
    tensors = read_tensor_network_file(filename)
    pathlib.Path(filename).unlink()

    # add the zero state tensor |0...0> at the beginning
    shape = [2] * nqubits
    data = sparse.COO(coords=[[0]] * nqubits, data=1, shape=shape, sorted=True, has_duplicates=False)
    inds = ["q" + str(i) + "_0" for i in range(nqubits)]
    tags = ["Q" + str(i) for i in range(nqubits)]
    tensors.insert(0, qtn.Tensor(data=data, inds=inds, tags=tags))

    # using the following lines instead would allow much greater flexibility,
    # but is not supported for DD-based simulation at the moment

    # add the zero state tensor |0>...|0> at the beginning
    # shape = [2]
    # data = np.zeros(2)
    # data[0] = 1.0
    # for i in range(qc.num_qubits):
    #     inds = ['q'+str(i)+'_0']
    #     tags = ['Q'+str(i)]
    #     tensors.insert(0, qtn.Tensor(data=data.reshape(shape), inds=inds, tags=tags))

    return qtn.TensorNetwork(tensors)


def get_simulation_path(
    qc: QuantumCircuit,
    max_time: int = 60,
    max_repeats: int = 1024,
    parallel_runs: int = 1,
    dump_path: bool = True,
    plot_ring: bool = False,
) -> list[tuple[int, int]]:
    """Determine a simulation path via computing a contraction path using cotengra.

    Args:
        qc: The quantum circuit to be simulated.
        max_time: The maximum time in seconds to spend on optimization.
        max_repeats: The maximum number of repetitions for optimization.
        parallel_runs: The number of parallel runs for optimization.
        dump_path: Whether to dump the path to a file.
        plot_ring: Whether to plot the contraction tree as a ring.

    Returns:
        The simulation path as a list of tuples.
    """
    import cotengra as ctg
    from opt_einsum.paths import linear_to_ssa

    tn = create_tensor_network(qc)

    opt = ctg.HyperOptimizer(
        max_time=max_time,
        max_repeats=max_repeats,
        progbar=True,
        parallel=parallel_runs,
        minimize="flops",
    )
    info = tn.contract(all, get="path-info", optimize=opt)
    path = cast(list[tuple[int, int]], linear_to_ssa(info.path))

    if dump_path:
        filename = qc.name + "_" + str(qc.num_qubits) + ".path" if isinstance(qc, QuantumCircuit) else "simulation.path"
        with pathlib.Path(filename).open("w", encoding=locale.getpreferredencoding(False)) as file:
            file.write(str(path))

    if plot_ring:
        fig, _ = opt.get_tree().plot_ring()
        fig.savefig("simulation_ring.svg", bbox_inches="tight")

    return path


class PathQasmSimulatorBackend(QasmSimulatorBackend):
    """Python interface to MQT DDSIM Simulation Path Framework."""

    _PATH_TARGET = Target(description="MQT DDSIM Simulation Path Framework Target", num_qubits=128)

    @staticmethod
    def _add_operations_to_target(target: Target) -> None:
        DDSIMTargetBuilder.add_0q_gates(target)
        DDSIMTargetBuilder.add_1q_gates(target)
        DDSIMTargetBuilder.add_2q_gates(target)
        DDSIMTargetBuilder.add_3q_gates(target)
        DDSIMTargetBuilder.add_multi_qubit_gates(target)
        DDSIMTargetBuilder.add_barrier(target)
        DDSIMTargetBuilder.add_measure(target)

    def __init__(
        self,
        name: str = "path_sim_qasm_simulator",
        description: str = "MQT DDSIM Simulation Path Framework",
    ) -> None:
        """Constructor for the DDSIM Simulation Path Framework QASM Simulator backend.

        Args:
            name: The name of the backend.
            description: The description of the backend.
        """
        super().__init__(name=name, description=description)

    @classmethod
    def _default_options(cls) -> Options:
        return Options(
            shots=None,
            parameter_binds=None,
            simulator_seed=None,
            pathsim_configuration=PathSimulatorConfiguration(),
            mode=None,
            bracket_size=None,
            alternating_start=None,
            gate_cost=None,
            seed=None,
            cotengra_max_time=60,
            cotengra_max_repeats=1024,
            cotengra_plot_ring=False,
            cotengra_dump_path=True,
        )

    @property
    def target(self) -> Target:
        """Return the target of the backend."""
        return self._PATH_TARGET

    def _run_experiment(self, qc: QuantumCircuit, **options: Any) -> ExperimentResult:
        start_time = time.time()

        pathsim_configuration = options.get("pathsim_configuration", PathSimulatorConfiguration())

        mode = options.get("mode")
        if mode is not None:
            pathsim_configuration.mode = PathSimulatorMode(mode)

        bracket_size = options.get("bracket_size")
        if bracket_size is not None:
            pathsim_configuration.bracket_size = bracket_size

        alternating_start = options.get("alternating_start")
        if alternating_start is not None:
            pathsim_configuration.alternating_start = alternating_start

        gate_cost = options.get("gate_cost")
        if gate_cost is not None:
            pathsim_configuration.gate_cost = gate_cost

        seed: int | None = options.get("seed")
        if seed is not None:
            pathsim_configuration.seed = seed

        sim = PathCircuitSimulator(qc, config=pathsim_configuration)

        # determine the contraction path using cotengra in case this is requested
        if pathsim_configuration.mode == PathSimulatorMode.cotengra:
            max_time = options.get("cotengra_max_time", 60)
            max_repeats = options.get("cotengra_max_repeats", 1024)
            dump_path = options.get("cotengra_dump_path", False)
            plot_ring = options.get("cotengra_plot_ring", False)
            path = get_simulation_path(
                qc,
                max_time=max_time,
                max_repeats=max_repeats,
                dump_path=dump_path,
                plot_ring=plot_ring,
            )
            sim.set_simulation_path(path, False)

        shots = options.get("shots", 1024)
        setup_time = time.time()
        counts = sim.simulate(shots)
        end_time = time.time()

        data = ExperimentResultData(
            counts={hex(int(result, 2)): count for result, count in counts.items()},
            statevector=None if not self._SHOW_STATE_VECTOR else sim.get_vector(),
            time_taken=end_time - start_time,
            time_setup=setup_time - start_time,
            time_sim=end_time - setup_time,
        )

        return ExperimentResult(
            shots=shots,
            success=True,
            status="DONE",
            config=pathsim_configuration,
            seed=seed,
            data=data,
            metadata=qc.metadata,
            header=DDSIMHeader(qc),
        )



"""Backend for DDSIM Hybrid Schrodinger-Feynman Simulator."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from qiskit import QuantumCircuit

from qiskit import QiskitError
from qiskit.providers import Options
from qiskit.result.models import ExperimentResult, ExperimentResultData
from qiskit.transpiler import Target
from qiskit.utils.multiprocessing import local_hardware_info

from .header import DDSIMHeader
from .pyddsim import HybridCircuitSimulator, HybridMode
from .qasmsimulator import QasmSimulatorBackend
from .target import DDSIMTargetBuilder


class HybridQasmSimulatorBackend(QasmSimulatorBackend):
    """Python interface to MQT DDSIM Hybrid Schrodinger-Feynman Simulator."""

    _HSF_TARGET = Target(description="MQT DDSIM HSF Simulator Target", num_qubits=128)

    @staticmethod
    def _add_operations_to_target(target: Target) -> None:
        DDSIMTargetBuilder.add_0q_gates(target)
        DDSIMTargetBuilder.add_1q_gates(target)
        DDSIMTargetBuilder.add_2q_controlled_gates(target)
        DDSIMTargetBuilder.add_barrier(target)
        DDSIMTargetBuilder.add_measure(target)

    def __init__(
        self,
        name: str = "hybrid_qasm_simulator",
        description: str = "MQT DDSIM Hybrid Schrodinger-Feynman simulator",
    ) -> None:
        """Constructor for the DDSIM Hybrid Schrodinger-Feynman simulator backend.

        Args:
            name: The name of the backend.
            description: The description of the backend.
        """
        super().__init__(name=name, description=description)

    @classmethod
    def _default_options(cls) -> Options:
        return Options(
            shots=None,
            parameter_binds=None,
            simulator_seed=None,
            mode="amplitude",
            nthreads=local_hardware_info()["cpus"],
        )

    @property
    def target(self) -> Target:
        """Return the target of the backend."""
        return self._HSF_TARGET

    def _run_experiment(self, qc: QuantumCircuit, **options: Any) -> ExperimentResult:
        start_time = time.time()
        seed = options.get("seed", -1)
        mode = options.get("mode", "amplitude")
        nthreads = int(options.get("nthreads", local_hardware_info()["cpus"]))
        if mode == "amplitude":
            hybrid_mode = HybridMode.amplitude
            max_qubits = 30  # hard-coded 16GiB memory limit
            algorithm_qubits = qc.num_qubits
            if algorithm_qubits > max_qubits:
                msg = "Not enough memory available to simulate the circuit even on a single thread"
                raise QiskitError(msg)
            qubit_diff = max_qubits - algorithm_qubits
            nthreads = int(min(2**qubit_diff, nthreads))
        elif mode == "dd":
            hybrid_mode = HybridMode.DD
        else:
            msg = f"Simulation mode{mode} not supported by hybrid simulator. Available modes are 'amplitude' and 'dd'."
            raise QiskitError(msg)

        sim = HybridCircuitSimulator(qc, seed=seed, mode=hybrid_mode, nthreads=nthreads)

        shots = options.get("shots", 1024)
        if self._SHOW_STATE_VECTOR and shots > 0:
            shots = 0

        counts = sim.simulate(shots)
        end_time = time.time()

        data = ExperimentResultData(
            counts={hex(int(result, 2)): count for result, count in counts.items()},
            statevector=None
            if not self._SHOW_STATE_VECTOR
            else sim.get_vector()
            if sim.get_mode() == HybridMode.DD
            else sim.get_final_amplitudes(),
            time_taken=end_time - start_time,
            mode=mode,
            nthreads=nthreads,
        )

        return ExperimentResult(
            shots=shots,
            success=True,
            status="DONE",
            seed=seed,
            data=data,
            metadata=qc.metadata,
            header=DDSIMHeader(qc),
        )
